import json
import os
import getpass
from socket import gethostname
from dataclasses import dataclass, fields
from tempfile import NamedTemporaryFile
from typing import List, Optional, Any

from celery import bootsteps
from celery.worker.components import Hub

from firexapp.application import get_app_tasks
from firexapp.common import silent_mkdir, create_link
from firexapp.submit.uid import FIREX_ID_REGEX
from firexkit.result import get_run_results_from_root_task_promise, RUN_RESULTS_NAME
from firexkit.task import convert_to_serializable
from celery.utils.log import get_task_logger
from firexapp.engine.celery import app

logger = get_task_logger(__name__)


@dataclass
class FireXRunData:
    completed: bool
    chain: List[str]
    firex_id: str
    logs_path: str
    submission_host: str
    submission_dir: str
    submission_cmd: List[str]
    viewers: dict[str, str]
    inputs: dict[str, Any]
    results: Optional[dict[str, Any]] = None
    revoked: bool = False

    def get_result(self, result_key, default=None):
        return (self.results or {}).get(RUN_RESULTS_NAME, {}).get(result_key, default)


def _get_common_run_data(uid, chain, submission_dir, argv, original_cli, inputs):
    if chain:
        chain = [t.short_name for t in get_app_tasks(chain)]
    if not argv:
        argv = []
    data = {
        **uid.run_data,
        'chain': chain,
        'logs_path': uid.logs_dir,
        'submission_host': app.conf.mc or gethostname(),
        'submission_dir': submission_dir,
        'submission_cmd': original_cli or list(argv),
        'submitter': getpass.getuser(),
        'inputs': inputs,
    }

    viewers = uid.viewers or {}
    data.update(viewers)
    data['viewers'] = viewers

    return data


class FireXJsonReportGenerator:
    formatters = ('json',)

    reporter_dirname = 'json_reporter'
    initial_report_filename = 'initial_report.json'
    completion_report_filename = 'completion_report.json'
    report_link_filename = 'run.json'

    @staticmethod
    def write_report_file(data, report_file):
        # Create the json_reporter dir if it doesn't exist
        silent_mkdir(os.path.dirname(report_file))

        # Atomic write, because the completed_run_json can be written from various places, including
        # celery poolworker which runs FireXRunner, celery mainprocess (as a last-resort backup in a bootstep),
        # and in another process (in the sync case). And although the backup method should kick in only after
        # other methods have failed, it's a theoretical possibility they will run concurrently depending
        # on the order of kill signals, especially in the sync case.
        with NamedTemporaryFile(mode='w', encoding='utf-8', dir=os.path.dirname(report_file), delete=False) as f:
            json.dump(convert_to_serializable(data),
                      fp=f,
                      skipkeys=True,
                      sort_keys=True,
                      indent=4)
            f.flush()
            os.fsync(f.fileno())

        os.chmod(f.name, 0o644)
        os.replace(f.name, report_file)

    @staticmethod
    def create_initial_run_json(uid, chain, submission_dir, argv, original_cli=None, json_file=None,
                                **inputs):
        data = _get_common_run_data(
            uid=uid,
            chain=chain,
            submission_dir=submission_dir,
            argv=argv,
            original_cli=original_cli,
            inputs=inputs) | {
                'completed': False,
                'revoked': False,
            }

        initial_report_file = get_initial_run_json_path(uid.logs_dir)
        FireXJsonReportGenerator.write_report_file(data, initial_report_file)

        report_link = os.path.join(uid.logs_dir, FireXJsonReportGenerator.report_link_filename)
        try:
            create_link(initial_report_file, report_link, delete_link=False)
        except FileExistsError:
            logger.debug(f'f{report_link} link already exist. '
                         f'No need to link to f{initial_report_file}')

        if json_file:
            try:
                create_link(report_link, json_file, delete_link=False, relative=True)
            except FileExistsError:
                logger.debug(f'{json_file} link already exist; '
                             f'post_run must have already created the link to {report_link}')

    @staticmethod
    def create_completed_run_json(uid=None, run_revoked=True, chain=None, root_id=None, submission_dir=None, argv=None,
                                  original_cli=None, json_file=None, logs_dir=None, **inputs):
        if not logs_dir and uid is None:
            raise ValueError('At least one of "logs_dir" or "uid" must be supplied')
        logs_dir = uid.logs_dir if uid is not None else logs_dir

        data = None
        try:
            with open(get_initial_run_json_path(logs_dir), encoding='utf-8') as f:
                data = json.load(fp=f)
        except OSError:
            logger.warning(f"Failed to read initial json for {logs_dir}. Creating a minimal completion report.")

        if not data and uid:
            # best effort -- not all termination contexts have access to all this data :/
            data = _get_common_run_data(
                uid=uid,
                chain=chain,
                submission_dir=submission_dir,
                argv=argv,
                original_cli=original_cli,
                inputs=inputs)

        data['completed'] = True
        data['results'] = get_run_results_from_root_task_promise(root_id)
        data['revoked'] = run_revoked

        completion_report_file = get_completion_run_json_path(logs_dir)
        FireXJsonReportGenerator.write_report_file(data, completion_report_file)

        report_link = os.path.join(logs_dir, FireXJsonReportGenerator.report_link_filename)
        create_link(completion_report_file, report_link, relative=True)

        if json_file:
            try:
                # This is typically not required, unless post_run ran before pre_run
                create_link(report_link, json_file, delete_link=False, relative=True)
            except FileExistsError:
                # This is expected for most cases
                pass


def get_completion_report_data(logs_dir):
    report_file = os.path.join(logs_dir, FireXJsonReportGenerator.reporter_dirname,
                               FireXJsonReportGenerator.completion_report_filename)
    with open(report_file) as f:
        return json.load(fp=f)


def is_completed_report(json_file: str) -> bool:
    return os.path.basename(os.path.realpath(json_file)) == FireXJsonReportGenerator.completion_report_filename


class RunJsonFileNotFound(AssertionError):
    pass


def load_completion_report(json_file: str) -> FireXRunData:
    if not os.path.isfile(json_file):
        raise RunJsonFileNotFound(f"File doesn't exist: {json_file}")

    with open(json_file) as f:
        run_dict = json.load(fp=f)

    field_names = {f.name for f in fields(FireXRunData)}
    filtered_run_dict = {k: v for k, v in run_dict.items() if k in field_names}
    # TODO: consider using dacite library instead.
    return FireXRunData(**filtered_run_dict)


def get_run_json_path(logs_dir: str) -> str:
    return os.path.join(
        logs_dir,
        FireXJsonReportGenerator.report_link_filename)


def get_initial_run_json_path(logs_dir):
    return os.path.join(
        logs_dir,
        FireXJsonReportGenerator.reporter_dirname,
        FireXJsonReportGenerator.initial_report_filename)


def get_completion_run_json_path(logs_dir):
    return os.path.join(
        logs_dir,
        FireXJsonReportGenerator.reporter_dirname,
        FireXJsonReportGenerator.completion_report_filename)


class ReporterStep(bootsteps.StartStopStep):

    def include_if(self, parent):
        return parent.hostname.startswith(app.conf.primary_worker_name + '@')

    def __init__(self, parent, **kwargs):
        self._logs_dir = None
        logfile = os.path.normpath(kwargs.get('logfile', '') or '')

        while (sp := os.path.split(logfile))[0] != logfile:
            m = FIREX_ID_REGEX.search(sp[1])
            if m:
                self._logs_dir = logfile
                break
            logfile = sp[0]

        super().__init__(parent, **kwargs)

    def stop(self, parent):
        if self._logs_dir and not os.path.exists(get_completion_run_json_path(logs_dir=self._logs_dir)):
            # By now, the report should have been written! Write a default completion report
            FireXJsonReportGenerator.create_completed_run_json(logs_dir=self._logs_dir)

app.steps['worker'].add(ReporterStep)

# We want this step to finish after Pool at least (because a poolworker writes this file in the async case),
# but might as well finish after Hub too
Hub.requires = Hub.requires + (ReporterStep,)



import traceback
from abc import ABC, abstractmethod
from functools import wraps

from celery.local import PromiseProxy
from celery.result import AsyncResult
from celery.states import SUCCESS
from celery.utils.log import get_task_logger

from firexkit.result import get_task_name_from_result
from firexkit.task import get_current_reports_uids

logger = get_task_logger(__name__)


REL_COMPLETION_REPORT_PATH = 'completion_email.html'

class ReportGenerator(ABC):
    formatters = tuple()
    loaders = tuple()

    @staticmethod
    def pre_run_report(**kwarg):
        """ This runs in the context of __main__ """
        pass

    @abstractmethod
    def add_entry(self, key_name, value, priority, formatters, **extra):
        pass

    @abstractmethod
    def load_data(self, key_name, value, loaders, **extra):
        pass

    @abstractmethod
    def post_run_report(self, root_id=None, **kwargs):
        """ This could runs in the context of __main__ if --sync, other in the context of celery.
            So the instance cannot be assumed be the same as in pre_run_report() """
        pass

    def filter_formatters(self, all_formatters):
        if not self.formatters:
            return all_formatters
        filtered_formatters = {f: all_formatters[f] for f in self.formatters if f in all_formatters}
        if not filtered_formatters:
            return None
        return filtered_formatters

    def filter_loaders(self, all_loaders):
        if not self.loaders:
            return all_loaders
        filtered_loaders = {l: all_loaders[l] for l in self.loaders if l in all_loaders}
        if not filtered_loaders:
            return None
        return filtered_loaders


class ReportersRegistry:
    _generators = None

    @classmethod
    def get_generators(cls):
        if not cls._generators:
            cls._generators = [c() for c in ReportGenerator.__subclasses__()]
        return cls._generators

    @classmethod
    def pre_run_report(cls, kwargs):
        for report_gen in cls.get_generators():
            report_gen.pre_run_report(**kwargs)

    @classmethod
    def post_run_report(cls, results, kwargs):
        if kwargs is None:
            kwargs = {}

        if results:
            from celery import current_app
            report_uids = get_current_reports_uids(current_app.backend)
            report_results = [AsyncResult(r, backend=current_app.backend) for r in report_uids]
            logger.debug(f"Processing reports for {report_uids}")
            for task_result in report_results:
                try:
                    # only report on successful tasks
                    if task_result.state != SUCCESS:
                        continue

                    task_name = get_task_name_from_result(task_result)
                    if task_name not in current_app.tasks:
                        continue

                    task = current_app.tasks[task_name]
                    report_entries = getattr(task, 'report_meta')

                    task_ret = task_result.result
                    for report_gen in cls.get_generators():
                        for report_entry in report_entries:
                            formatters = report_entry.get("formatters", [])
                            loaders = report_entry.get("loaders", [])
                            key_name = report_entry["key_name"]
                            logger.debug(f"Processing report entry for task {task_name} with key_name {key_name}")
                            if len(loaders) > 0:
                                logger.debug(f'Loading report data for task {task_name}')
                                filtered_loaders = report_gen.filter_loaders(loaders)

                                if filtered_loaders is None:
                                    continue

                                try:
                                    report_gen.load_data(
                                        key_name=key_name,
                                        value=task_ret[key_name] if key_name else task_ret,
                                        loaders=filtered_loaders,
                                        all_task_returns=task_ret,
                                        task_name=task_name,
                                        task_uuid=task_result.id
                                    )
                                    logger.debug(f'Completed loading report data for task {task_name}')
                                except Exception:
                                    logger.error(f'Error during report data loading for task {task_name}...skipping', exc_info=True)
                                    continue
                            if len(formatters) > 0:
                                logger.debug(f'Adding report entry for task {task_name}')
                                filtered_formatters = report_gen.filter_formatters(formatters)

                                if filtered_formatters is None:
                                    continue

                                try:
                                    report_gen.add_entry(
                                        key_name=key_name,
                                        value=task_ret[key_name] if key_name else task_ret,
                                        priority=report_entry["priority"],
                                        formatters=filtered_formatters,
                                        all_task_returns=task_ret,
                                        task_name=task_name,
                                        task_uuid=task_result.id)
                                    logger.debug(f'Completed adding report entry for task {task_name}')
                                except Exception:
                                    logger.error(f'Error during report generation for task {task_name}...skipping', exc_info=True)
                                    continue

                except Exception:
                    logger.error(f"Failed to add report entry for task result {task_result}", exc_info=True)

            logger.debug("Completed processing results data for reports")

        for report_gen in cls.get_generators():
            try:
                logger.debug(f'Running post_run_report for {report_gen}')
                report_gen.post_run_report(root_id=results, **kwargs)
                logger.debug(f'Completed post_run_report for {report_gen}')
            except Exception:
                # Failure in one report generator should not impact another
                logger.error(f'Error in the post_run_report for {report_gen}', exc_info=True)


def report(key_name=None, priority=-1, **formatters):
    """ Use this decorator to indicate what returns to include in the report and how to format it """

    def decorator(func):

        @wraps(func)
        def tag_with_report_meta_data(cls):

            # guard: prevent bad coding by catching bad return key
            if key_name and key_name not in cls.return_keys:
                raise Exception("Task %s does not specify %s using the @returns decorator. "
                                "It cannot be used in @report" % (cls.name, key_name))

            report_entry = {
                "key_name": key_name,
                'priority': priority,
                'formatters': formatters,
            }
            if not cls.has_report_meta():
                cls.report_meta = []
            cls.report_meta.append(report_entry)
            return cls

        if hasattr(func, '__qualname__'):
            return tag_with_report_meta_data(func)

        logger.debug(f"Skipping applying @report since {func} is not a PromiseProxy type")

        return func

    return decorator


def report_data(key_name=None, **loaders):
    """ Use this decorator to indicate what returns to include in the report and how to load it """

    def decorator(func):

        @wraps(func)
        def tag_with_report_meta_data(cls):

            # guard: prevent bad coding by catching bad return key
            if key_name and key_name not in cls.return_keys:
                raise Exception("Task %s does not specify %s using the @returns decorator. "
                                "It cannot be used in @report" % (cls.name, key_name))

            report_data = {
                "key_name": key_name,
                'loaders': loaders,
            }
            if not cls.has_report_meta():
                cls.report_meta = []
            cls.report_meta.append(report_data)
            return cls

        if hasattr(func, '__qualname__'):
            return tag_with_report_meta_data(func)

        logger.debug(f"Skipping applying @report_data since {func} is not a PromiseProxy type")

        return func

    return decorator


import datetime
from dataclasses import dataclass
from enum import Enum, auto
import logging
import os
import shlex
import shutil
from socket import gethostname
import stat
import subprocess
import tempfile
import time
import urllib.parse
import uuid
from typing import Union, Optional
import glob
import psutil
from celery.utils.log import get_task_logger

from firexkit import firex_exceptions, firexkit_common
from firexapp.engine.logging import html_escape
from firexapp.events.model import EXTERNAL_COMMANDS_KEY

logger = get_task_logger(__name__)


@dataclass
class ProcStats:
    include_children: bool = True  # Input: include child processes (recursively). Note that if the spawned
                                   # process disowns its children, we will not be able to capture those

    collection_interval: int = 2  # Input: Seconds between stats sampling. Sampling isn't free, so
                                  # best not to be too aggressive. Note however that if the process
                                  # spawns children which rapidly blink in and out of existence,
                                  # the default value will not capture much at all

    interval_increase: bool = True # Input: Increase collection interval by elapsed_time * collection_interval/120
                                   # up to a maximum of collection_interval * 30 seconds

    cpu_percent_used: int = 0 # CPU percent used over the time period
    mem_mb_used: int = 0  # Average of memory usage over time in MiB (pss if available, else rss)
    mem_mb_high_wm: int = 0  # High watermark for memory usage
    elapsed_time: float = 0.0  # In seconds
    num_cpu: int = 0
    total_mem : int = 0  # in Mib

    def clear_stats(self):
        """
        Clears the stats (but not the input knobs)
        """
        self.cpu_percent_used = 0
        self.mem_mb_used = 0
        self.mem_mb_high_wm = 0
        self.elapsed_time = 0.0
        self.num_cpu = 0
        self.total_mem = 0


class CommandFailed(firex_exceptions.FireXCalledProcessError):
    pass


class _SubprocessRunnerType(Enum):
    CHECK_OUTPUT = auto()
    CHECK_CALL = auto()
    RUN = auto()


def open_og_rw_permissions(path):
    rw_group = stat.S_IRGRP | stat.S_IWGRP
    rw_other = stat.S_IROTH | stat.S_IWOTH
    os.chmod(path, os.stat(path).st_mode | rw_group | rw_other)


def check_call(cmd, retries=0, retry_delay=3, **kwargs):
    runner_type = _SubprocessRunnerType.CHECK_CALL

    _sanitize_runner_kwargs(runner_type=runner_type, kwargs=kwargs)
    _subprocess_runner_retries(runner_type=runner_type, capture_output=False, check=True,
                               retries=retries, retry_delay=retry_delay, cmd=cmd, **kwargs)


def check_output(cmd, retries=0, retry_delay=3, **kwargs):
    runner_type = _SubprocessRunnerType.CHECK_OUTPUT

    _sanitize_runner_kwargs(runner_type=runner_type, kwargs=kwargs)
    result = _subprocess_runner_retries(runner_type=runner_type, capture_output=True, check=True,
                                        retries=retries, retry_delay=retry_delay, cmd=cmd, **kwargs)
    return result.stdout


def run(cmd, retries=0, retry_delay=3, **kwargs):
    # Note that this differs from Python 3.8's API since capture_output is True by default in _subprocess_runner

    runner_type = _SubprocessRunnerType.RUN

    _sanitize_runner_kwargs(runner_type=runner_type, kwargs=kwargs)
    return _subprocess_runner_retries(runner_type=runner_type,  # Pass through 'capture_output' and 'check'
                                      retries=retries, retry_delay=retry_delay, cmd=cmd, **kwargs)


def _sanitize_runner_kwargs(runner_type: _SubprocessRunnerType, kwargs: dict):
    disallowed_keys = ['stdout', 'universal_newlines', 'text']
    if runner_type is _SubprocessRunnerType.RUN:
        disallowed_keys.extend(['input'])
    else:
        disallowed_keys.extend(['capture_output', 'check'])

    for key in disallowed_keys:
        if key in kwargs:
            logger.error('[%s] WARNING: %s argument not allowed, it will be overridden.' %
                         (runner_type.name.lower(), key))
            del kwargs[key]


def _subprocess_runner_retries(retries, retry_delay, **kwargs):
    if retries is None:
        retries = 0
    if retry_delay is None:
        retry_delay = 0
    retry_count = 0
    while True:
        try:
            extra_header = '[retry %d/%d]' % (retry_count, retries) if retry_count else None
            return _subprocess_runner(extra_header=extra_header, **kwargs)
        except (subprocess.SubprocessError, FileNotFoundError):
            if retry_count >= retries:
                raise
            else:
                retry_count += 1
                time.sleep(retry_delay)


def _send_flame_subprocess(subprocess_data):
    try:
        from celery import current_task
        if current_task:
            current_task.send_firex_event_raw({EXTERNAL_COMMANDS_KEY: subprocess_data})
    except Exception as e:
        logger.warning(f"Error while sending flame subprocess event: {e}")


def _send_flame_subprocess_start(flame_subprocess_id, cmd, cwd, host, filename=None, remote_host=None):
    _send_flame_subprocess({flame_subprocess_id: {'cmd': cmd,
                                                  'cwd': str(cwd) if cwd else cwd,
                                                  'output_file': filename,
                                                  'host': host,
                                                  'remote_host': remote_host,
                                                  'start_time': time.time()}})


def _send_flame_subprocess_end(flame_subprocess_id, output, returncode, chars=None, hung_process=None,
                               slow_process=None, stderr_output=None):
    ui_max_chars = 8000
    chars = ui_max_chars if chars is None else min(chars, ui_max_chars)
    subproc_result = {
        'completed': not hung_process and not slow_process and returncode is not None,
        'timeout': slow_process,
        'inactive': hung_process,
        'output_truncated': len(output) >= chars if output is not None else False,
        'output': output[-chars:] if output else output,
        'stderr_output_truncated': len(stderr_output) >= chars if stderr_output is not None else False,
        'stderr_output': stderr_output[-chars:] if stderr_output else stderr_output,
        'returncode': returncode,
    }
    _send_flame_subprocess({flame_subprocess_id: {'result': subproc_result, 'end_time': time.time()}})


def _subprocess_runner(cmd: Union[str, list], runner_type: _SubprocessRunnerType = _SubprocessRunnerType.CHECK_OUTPUT,
                       extra_header=None, file=None, chars=32000, timeout=None, capture_output=True, check=False,
                       inactivity_timeout=30 * 60, monitor_activity_files=None, log_level=logging.DEBUG,
                       copy_file_path=None, shell=False, cwd=None, env=None, remove_firex_pythonpath=True,
                       logger=logger, stderr=subprocess.STDOUT, proc_stats: Optional[ProcStats] = None,
                       stdin=subprocess.PIPE, bufsize=0, **kwargs):
    ##########################
    # Local Helper functions #
    ##########################
    def _get_cmd_str():
        if isinstance(cmd, list):
            return ' '.join([shlex.quote(token) for token in cmd])
        else:
            return cmd

    def _sanitize_cmd():
        nonlocal cmd
        if isinstance(cmd, str) and not shell:
            cmd = shlex.split(cmd)

    def _get_env_without_pythonpath(user_env):
        try:
            current_env_pythonpath = os.environ['PYTHONPATH']
        except KeyError:
            # if we don't have a PYTHONPATH in the current env, then we can't do anything
            return user_env
        else:
            if user_env is None:
                # User didn't supply an env, so, we remove our PYTHONPATH
                current_env_copy = os.environ.copy()
                del current_env_copy['PYTHONPATH']
                return current_env_copy
            else:
                # User provided a custom env
                try:
                    user_env_pythonpath = user_env['PYTHONPATH']
                except KeyError:
                    # If user provided custom env didn't have PYTHONPATH, then use as-is
                    return user_env
                else:
                    # Remove our PYTHONPATH from the user's env PYTHONPATH
                    current_env_pythonpath_items = current_env_pythonpath.split(':')
                    user_env_pythonpath_items = user_env_pythonpath.split(':')
                    new_pythonpath = [x for x in user_env_pythonpath_items if (x not in current_env_pythonpath_items)]

                    user_env_copy = dict(user_env)
                    user_env_copy['PYTHONPATH'] = ':'.join(new_pythonpath)
                    return user_env_copy

    def _log_intro_msg(subprocess_uuid):
        msg = [f'{log_header} Executing the following command:']
        sep = '$' if shell else '>'
        live_link = _get_live_file_monitor_link()
        if live_link is None:
            live_link = ''
        msg += [f'{live_link}'
                f'<span class="command_line_prefix">{host}:{cwd_str}{sep}</span> '
                f'<span class="command_line">{html_escape(cmd_str)}</span>']
        span_class = 'command_extra_info' if file else 'hidden'
        msg += [f'<span class="{span_class}">output also written to: {filename}</span>']
        logger.log(log_level, '\n'.join(msg), extra={'label': subprocess_uuid,
                                                     'span_class': 'command',
                                                     'html_escape': False})

    def _get_live_file_monitor_link():
        from firexapp.engine.celery import app
        try:
            # FIXME: https://github.com/FireXStuff/firexapp/issues/10
            if app.conf.install_config.has_viewer():
                run_url = app.conf.install_config.get_run_url()
            else:
                return
        except AttributeError:
            return

        # Note we can't use urllib.parse.urljoin because the path is in the fragment, which urljoin can only overwrite.
        # FIXME: ideally flame-ui URL path knowledge would be externalized.
        if not run_url.endswith('/'):
            run_url += '/'
        live_link_url = run_url + f"live-file?file={urllib.parse.quote(filename, safe='')}&host={host}"
        link = firexkit_common.get_link(live_link_url,
                                        text='',
                                        title_attribute=f'Live monitor of {filename}',
                                        attrs={"target": "_blank", "style": "margin-right:5px"},
                                        other_elements='<i class="far fa-eye"></i>',
                                        html_class=live_file_monitor_span_class)
        return link

    def _hide_live_file_monitor_element():
        logger.raw('<style>.%s {display: none;}</style>' % live_file_monitor_span_class)

    def _get_output_from_file():
        if not file or not chars or os.fstat(f.fileno()).st_size < chars:
            f.seek(0)
        else:
            f.seek(-chars, 2)

        return f.read().decode(encoding='utf-8', errors='replace')

    def _log_output(contents, error=False):
        if not chars or len(contents) < chars:
            log_msg = '%s Returned String:\n%s' % (log_header, contents)
        else:
            log_msg = '%s Last %d chars of Returned String:\n%s' % (log_header, chars, contents[-chars:])
        logger.log(log_level,
                   log_msg,
                   extra={'span_class': 'command_output command_output_error' if error else 'command_output'})

    def _kill_proc_gently(proc):
        children = None
        try:
            children = proc.children(recursive=True)
            proc.terminate()
            try:
                proc.wait(timeout=60)
            except (subprocess.TimeoutExpired, psutil.TimeoutExpired):
                # Okay, kill not so gently
                proc.kill()
                try:
                    proc.wait(timeout=6)
                except (subprocess.TimeoutExpired, psutil.TimeoutExpired) as e:
                    # Give up at this point. It is undead.
                    logger.exception(e)
                    pass
        except (PermissionError, psutil.AccessDenied, psutil.NoSuchProcess) as e:
            # Possible if the underlying process is running under sudo or the like
            logger.exception(e)

        if not children:
            return #  <- we are done

        # check if the process left children around after it had a chance to clean up, and kill those
        # NOTE: if process spawned a child between children() and terminate() above, we are out of luck
        _, alive = psutil.wait_procs(children, timeout=0)
        if not alive:
            return  # <- we are done

        # first pass terminate
        for live in alive:
            try:
                live.terminate()
            except (psutil.AccessDenied, psutil.NoSuchProcess):
                pass

        _, alive = psutil.wait_procs(alive, timeout=60)
        if not alive:
            return  # <- we are done

        # second pass kill
        for live in alive:
            try:
                live.kill()
            except (psutil.AccessDenied, psutil.NoSuchProcess):
                pass

        # third pass notify
        _, alive = psutil.wait_procs(alive, timeout=6)
        if not alive:
            return  # <- we are done

        logger.error(f'Failed to kill children of killed process. Children: {alive}')

    _cpu_stat_errors = set()
    _children_errors = set()
    _mem_stat_errors = set()
    _cpu_totals_by_proc = {}

    def _get_proc_cpu_totals(the_proc: psutil.Process)-> Union[float, None]:
        try:
            cpu_times = the_proc.cpu_times()
        except psutil.NoSuchProcess:
            # expected condition
            return None  # <-- Done. Note, return None, not 0! Otherwise, we would overwrite previous numbers

        except psutil.AccessDenied:
            if the_proc not in _cpu_stat_errors:  # Don't flood the console with debugs
                _cpu_stat_errors.add(the_proc)
                logger.debug(f'Cannot get cpu stats for {the_proc}')

            return None  # <-- Done

        return cpu_times.user + cpu_times.system

    def _get_proc_mem_totals(the_proc: psutil.Process)-> float:
        try:
            mem_info = the_proc.memory_full_info()
        except psutil.NoSuchProcess:
            # expected condition
            return 0.0  # <-- Done

        except psutil.AccessDenied:
            if the_proc not in _mem_stat_errors:  # Don't flood the console with debugs
                _mem_stat_errors.add(the_proc)
                logger.debug(f'Cannot get memory stats for {the_proc}')

            return 0.0  # <-- Done

        return getattr(mem_info, 'pss', mem_info.rss) / (1024.0 * 1024.0)

    # should only be called if we have proc_stats
    def _collect_stats():
        elapsed_time = now - start_time
        if not elapsed_time:
            return  # <-- Time is at a standstill; nothing to do

        proc_stats.elapsed_time = elapsed_time
        mem_totals = 0.0

        if proc_stats.include_children:
            try:
                children = p.children(recursive=True)
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                if p not in _children_errors:  # Don't flood the console with debugs
                    _children_errors.add(p)
                    logger.debug(f'Cannot get child processes for {p}', exc_info=True)

            else:  # p.children() successful
                for child_p in reversed(children):
                    with child_p.oneshot():
                        if (cpu := _get_proc_cpu_totals(child_p)) is not None:
                            _cpu_totals_by_proc[child_p] = cpu
                        mem_totals += _get_proc_mem_totals(child_p)

        with p.oneshot():
            if (cpu := _get_proc_cpu_totals(p)) is not None:
                _cpu_totals_by_proc[p] = cpu
            mem_totals += _get_proc_mem_totals(p)

        mem_totals = round(mem_totals)
        if mem_totals > proc_stats.mem_mb_high_wm:
            proc_stats.mem_mb_high_wm = mem_totals

        # we do a time-weighted average for memory
        current_weight = (now - last_proc_stats) / elapsed_time
        proc_stats.mem_mb_used = round(((1 - current_weight) * proc_stats.mem_mb_used) + (current_weight * mem_totals))
        # CPU totals will be calculated once after the main loop

    def _find_matching_files(paths_and_patterns, work_dir):
        if paths_and_patterns is None: return []
        if isinstance(paths_and_patterns, str):
            paths_and_patterns = [paths_and_patterns]
        matching_files = []
        for item in paths_and_patterns:
            item_path = os.path.join(work_dir, item) if not os.path.isabs(item) else item
            logger.info(f"Searching for: {item_path}")
            if os.path.isfile(item_path):
                matching_files.append(item_path)
            else:
                glob_matches = glob.glob(item_path, recursive=True)
                files_only = [match for match in glob_matches if os.path.isfile(match)]
                matching_files.extend(files_only)
        return matching_files

    def _get_size_of_files(files, work_dir):
        files_size = 0
        found_files = _find_matching_files(files, work_dir)
        if found_files:
            logger.debug(f"Checking for activity in: {found_files}")
        for found_file in found_files:
            try:
                files_size += os.stat(found_file).st_size
            except OSError as e:
                logger.error(f"An error occurred while reading the size of file '{found_file}': {e}")
        return files_size

    #################
    # Start of code #
    #################
    log_header = f'[{runner_type.name.lower()}]'
    log_header += extra_header if extra_header else ''
    cmd_str = _get_cmd_str()
    cwd_str = cwd if cwd is not None else os.getcwd()
    host = gethostname()
    subprocess_id = str(uuid.uuid4())
    live_file_monitor_span_class = f'live_file_{subprocess_id}'

    _sanitize_cmd()
    if remove_firex_pythonpath:
        env = _get_env_without_pythonpath(env)

    if file:
        f = open(file, 'wb+', buffering=0)
    else:
        f = tempfile.NamedTemporaryFile(delete=False, buffering=0)
    filename = f.name
    open_og_rw_permissions(filename)


    if log_level is not None:
        _send_flame_subprocess_start(flame_subprocess_id=subprocess_id, cmd=cmd, filename=filename, cwd=cwd_str,
                                     host=host)
        _log_intro_msg(subprocess_id)

    if proc_stats is not None:
        proc_stats.clear_stats()
        proc_stats.num_cpu = psutil.cpu_count() or 1
        proc_stats.total_mem = psutil.virtual_memory().total // (1024*1024)
        stats_interval = proc_stats.collection_interval

    slow_process = hung_process = False
    with f:
        p = output = None
        try:
            # Run the command
            p = psutil.Popen(cmd, stdin=stdin, stdout=f, stderr=stderr, shell=shell, cwd=cwd,
                             env=env, bufsize=bufsize, **kwargs)

            start_time = time.monotonic()

            last_output_size = 0
            last_file_output_size = current_file_output_size = 0
            last_proc_stats = last_log_time = last_output_time = last_file_size_check = start_time
            last_output_clock_time = time.time()
            _sleep = 0.05
            # Wait for process to finish
            while True:
                try:
                    p.wait(_sleep)
                    # If we get here, process is done
                    break
                except (subprocess.TimeoutExpired, psutil.TimeoutExpired):
                    pass
                _sleep = _sleep * 1.1 if _sleep * 1.1 < 1 else 1  # Exponential backoff

                now = time.monotonic()

                # Stats
                if proc_stats is not None and (now - last_proc_stats) >= stats_interval:
                    _collect_stats()
                    last_proc_stats = now

                    # Interval-increase calculation
                    if proc_stats.interval_increase:
                        max_stats_interval = proc_stats.collection_interval * 30.0
                        if stats_interval < max_stats_interval:
                            stats_interval = proc_stats.collection_interval + ((now - start_time) *
                                                                               proc_stats.collection_interval / 120)
                            if stats_interval > max_stats_interval:
                                stats_interval = max_stats_interval

                # Log
                if now - last_log_time > 10 * 60:  # Log every 10 minutes
                    time_str = datetime.datetime.fromtimestamp(last_output_clock_time).strftime('%Y-%m-%d %H:%M:%S')
                    mfs = f', and monitored files are {last_file_output_size} bytes.' if monitor_activity_files else '.'
                    logger.info(f'Waiting for command to finish...\n(Last command output was at {time_str}. '
                                f'Output size is {last_output_size} bytes{mfs}')
                    last_log_time = now

                # Hard timeout check
                if timeout and now - start_time > timeout:
                    # Process too slow. Kill it.
                    logger.debug(f'Total timeout {timeout} exceeded, killing pid {p.pid}')
                    _kill_proc_gently(p)
                    slow_process = True
                    break

                # Inactivity timeout check
                current_output_size = os.fstat(f.fileno()).st_size
                activity_timeout = inactivity_timeout and now - last_output_time > inactivity_timeout
                if monitor_activity_files and (now - last_file_size_check > 5 * 60 or activity_timeout):
                    # check size every 5 minutes, or just before we time out:
                    current_file_output_size = _get_size_of_files(monitor_activity_files, work_dir=cwd_str)
                    last_file_size_check = now

                if last_output_size != current_output_size or last_file_output_size != current_file_output_size:
                    # We have some activity!
                    last_output_size = current_output_size
                    last_file_output_size = current_file_output_size
                    last_output_time = now
                    last_output_clock_time = time.time()
                elif activity_timeout:
                    # Process hung. Kill it.
                    logger.debug(f'Activity (writing to stdout/stderr or monitored file) timeout {inactivity_timeout}'
                                 f' exceeded, killing pid {p.pid}')
                    _kill_proc_gently(p)
                    hung_process = True
                    break
            # End wait-for-process loop

            if proc_stats is not None and proc_stats.elapsed_time:
                proc_stats.cpu_percent_used = round(100 * sum(_cpu_totals_by_proc.values()) /
                                                    (proc_stats.elapsed_time * proc_stats.num_cpu))

            if capture_output:
                output = _get_output_from_file()
                if log_level is not None:
                    _log_output(output, error=True if p.returncode else False)

            # Should these exceptions be thrown after all the cleanups below?
            if slow_process:
                raise subprocess.TimeoutExpired(cmd, timeout, output=output, stderr=output)
            elif hung_process:
                raise firex_exceptions.FireXInactivityTimeoutExpired(cmd, inactivity_timeout, output=output,
                                                                     stderr=output)
        finally:
            if log_level is not None:
                _send_flame_subprocess_end(flame_subprocess_id=subprocess_id, hung_process=hung_process,
                                           slow_process=slow_process, output=output, chars=chars,
                                           returncode=getattr(p, 'returncode', None))
                _hide_live_file_monitor_element()

            if p and p.stdin:
                p.stdin.close()

            # Copy output file
            if (copy_file_path is not None and
                    not os.path.isfile(copy_file_path) and
                    os.path.isfile(filename)):
                shutil.copyfile(filename, copy_file_path)

    # Attempt to remove the temporary file
    if not file:
        try:
            os.remove(filename)
        except FileNotFoundError:
            logger.error('Could not delete temp file: %s' % filename)

    # Raise exception on error, if requested
    if check and p.returncode:
        raise CommandFailed(p.returncode, p.args, output=output, stderr=output)

    return subprocess.CompletedProcess(p.args, p.returncode, stdout=output, stderr=output)
